fs = require('fs')
coffee = require('coffee-script')

K = require('kcore')

COMPACT = true
# hard LB - always present
LBH = '\n'
# soft LB - present only when COMPACT = false
LB = if COMPACT then '' else '\n'

# turn off the cache when you are rapidly developing js/css etc.
RESOURCE_CACHE_ON = true
RESOURCE_CACHE = {}
RESOURCE_COMPILED_CACHE = {}

class K.JsCollector
  constructor: () ->
    @js = ''

  _generateJsTagFromText: (jsTxt) ->
    return '<script language="javascript" type="text/javascript"><!--' + LBH + jsTxt + LBH + '--></script>' + LB

  getAllAsTag: () ->
    rv = ''
    if @js != ''
      rv = @_generateJsTagFromText(@js)
    return rv

  getAllAsOnLoadCode: () ->
    rv = ''
    if @js != ''
      onLoadJs = '$(function() {' + LB + @js + LB + '});'
      rv = @_generateJsTagFromText(onLoadJs)
    return rv

  pushFromTxt: (txt) ->
    @js += txt.trim()

class K.NateHtml
  constructor: (parent) ->
    @children = []

  delete: () ->
    ; # TODO: implement this, but at the moment we need it only in K.NateHtmlDom

  newNate: (elem = null) ->
    return new K.NateHtml(elem)

  push: (child) ->
    @children.push child

  render: () ->
    txt = ''
    for child in @children
      txt += child.render()
    return txt

  @setResourceCacheEnabled: (onOff) ->
    RESOURCE_CACHE_ON = onOff

class K.NateHtmlHead extends K.NateHtml
  constructor: (parent) ->
    super(parent)
    @htmlDocument = parent
    @css = ''
    @meta = {}
    @alternateLinksForLanguages = null

  setTitle: (title) ->
    @meta.title = title

  setMetaAuthor: (author) ->
    @meta.author = author

  setMetaDescription: (description) ->
    @meta.description = description

  setMetaKeywords: (keywords) ->
    @meta.keywords = keywords

  setLinksForAlternateLangauges: (links) ->
    @alternateLinksForLanguages = links

  render: () ->
    txt = '<head>' + LB

    # title and metas
    txt += '<meta charset="UTF-8">' + LB
    if @meta.title?
      txt += '<title>' + @meta.title + '</title>' + LB
    if @meta.description?
      txt += '<meta name="description" content="' + @meta.description + '">' + LB
    if @meta.keywords?
      txt += '<meta name="keywords" content="' + @meta.keywords + '">' + LB
    if @meta.author?
      txt += '<meta name="author" content="' + @meta.author + '">' + LB
    txt += '<meta http-equiv="Cache-Control" content="no-cache"/>' + LB
    txt += '<meta http-equiv="Expires" content="0"/>' + LB
    txt += '<meta http-equiv="X-UA-Compatible" content="IE=Edge"/>' + LB
    txt += '<meta name="viewport" content="width=device-width, initial-scale=1">' + LB

    # alternate links for different langauges
    if @alternateLinksForLanguages?
      for langId, url of @alternateLinksForLanguages
        txt += '<link rel="alternate" hreflang="' + langId + '" href="' + url + '" />' + LB

    # css and js
    if @css != ''
      txt += '<style type="text/css">' + LB + @css + LB + '</style>' + LB
    txt += @htmlDocument.jsGetCollector('head').getAllAsTag()

    # finish
    txt += '</head>' + LB
    return txt

  pushCss: (css) ->
    @css += css.trim() + LB

class K.NateHtmlDocument extends K.NateHtml
  constructor: (parent) ->
    super(parent)
    @theHead = new K.NateHtmlHead(@)
    @theBody = new K.NateHtmlBody(@)
    @jsPipes =
      head:    new K.JsCollector()
      bodyEnd: new K.JsCollector()
      onLoad:  new K.JsCollector()
    @resourcesToLoad =
      css:        []
      js_head:    []
      js_bodyEnd: []
      js_onLoad:  []
    @resourcesToLoadCnt = 0
    @resourcesLoadDoneCb = null
    @waitingToEmit = false
    @langId = null

    # debugging helpers
    if 0
      @LOG_resourceLoading = console.log
    else
      @LOG_resourceLoading = () -> ;

  head: () ->
    return @theHead

  body: () ->
    return @theBody

  render: () ->
    if @langId?
      langMeta = ' lang="' + @langId + '"'
    else
      langMeta = ''

    txt = '<!DOCTYPE HTML>' + LB
    txt += '<html' + langMeta + '>' + LB
    txt += @theHead.render()
    txt += @theBody.render()
    txt += '</html>' + LB
    return txt

  setLang: (langId) ->
    @langId = langId

  loadResource: (resourceData, done) ->
    if RESOURCE_CACHE_ON and RESOURCE_CACHE[resourceData.path]?
      @LOG_resourceLoading '[Nate] RESOURCE_CACHE hit !'
      done(RESOURCE_CACHE[resourceData.path])
    else
      @LOG_resourceLoading '[Nate] RESOURCE_CACHE miss !'
      fs.readFile resourceData.path, 'utf8', (err, data) =>
        if not err?
          RESOURCE_CACHE[resourceData.path] = data
          done(data)
        else
          console.log err
          throw new Error('File loading failed from path:' + resourceData.path)

  # important:
  #   done() callback is not called at the moment the resource is loaded as this may cause change of order
  #   (for example, loading: A-B-C, then calling done(): A-C-B, may cause a lot of problems)
  #   done() is called in right order after ALL resources are loaded
  addResourceToLoad: (pipe, resourceData, done) ->
    # add load request to loading-pipe
    @resourcesToLoadCnt++
    resourceData.loaded = false
    resourceData.done   = done
    @resourcesToLoad[pipe].push resourceData
    @LOG_resourceLoading '[Nate] add resource:', resourceData.path

    # get from cache or trigger loading resource
    @loadResource resourceData, (data) =>
      resourceData.loaded = true
      @resourcesToLoadCnt--
      @LOG_resourceLoading '[Nate] loaded resource:', resourceData.path
      resourceData.loadedContent = data
      @_resourcesCheckForResourcesReady()

  _resourcesCheckForResourcesReady: () ->
    if (@waitingToEmit) and (@resourcesToLoadCnt == 0)
      if @resourcesLoadDoneCb?
        @_resourcesLoadDoneCb()
        @_resourcesLoadDoneCb = null

  _resourcesLoadDoneCb: () ->
    @LOG_resourceLoading '[Nate] All resources loaded...'
    for resourceType of @resourcesToLoad
      @resourcesToLoadItem = @resourcesToLoad[resourceType]
      for resourceItem in @resourcesToLoadItem
        resourceItem.done(resourceItem.loadedContent)
    @resourcesLoadDoneCb?()

  cssFromStaticFile: (path, done) ->
    @addResourceToLoad 'css', {path:path}, (data) =>
      @theHead.pushCss(data)
      done?()

  # supported pipes are:
  #  head - goes to <head></head>
  #  bodyEnd - goes to bottom of <body></body>
  #  onLoad - goes to jQuery onLoad handler at the end of <body></body> section
  jsGetCollector: (pipeName) ->
    return @jsPipes[pipeName]

  jsFromStaticFile: (path, pipeName, done) ->
    K.Error.reportIfParameterNotSet(path, 'javascript path')
    @addResourceToLoad 'js_' + pipeName, {path:path}, (data) =>
      @jsGetCollector(pipeName).pushFromTxt(data)
      done?()

  coffeeFromStaticFile: (path, pipeName, done) ->
    K.Error.reportIfParameterNotSet(path, 'coffeeScript path')
    @addResourceToLoad 'js_' + pipeName, {path:path}, (data) =>
      # compile or get from cache
      if RESOURCE_CACHE_ON and RESOURCE_COMPILED_CACHE[path]?
        compiled = RESOURCE_COMPILED_CACHE[path]
      else
        compiled = coffee.compile(data)
        RESOURCE_COMPILED_CACHE[path] = compiled
      # emit and done()
      @jsGetCollector(pipeName).pushFromTxt(compiled)
      done?()

  _onReadyToEmit: (done) ->
    @waitingToEmit = true
    @resourcesLoadDoneCb = done
    @_resourcesCheckForResourcesReady()

  emitAsResponse: (httpResponse, cb) ->
    @_onReadyToEmit () =>
      httpResponse.send @render()
      cb?()

  emitToString: (cb) ->
    @_onReadyToEmit () =>
      cb?(@render())

class K.NateHtmlElem extends K.NateHtml
  @NOTHING: 0
  @JUST_CHILDREN: 1
  @JUST_THE_INNER: 2
  @OPEN_CLOSE_AT_ONCE: 3
  @OPEN_THEN_CLOSE: 4
  @ON_CLICK_EVENT_BLOCKER: 'var event=arguments[0] || window.event; var rv = (event.button != 0); event.returnValue = rv; return rv;'

  constructor: (parent) ->
    super(parent)
    if parent?
      parent.push(@)
    else
      @tagOpenCloseMode = K.NateHtmlElem.JUST_CHILDREN
    @tagParamsList = {}
    @enabled = true

  deleteAllChildren: () ->
    ; # TODO: implement this, but at the moment we need it only in K.NateHtmlDom

  newNate: (parent) ->
    return new K.NateHtmlElem(parent)

  setTag: (@tagName, @tagParamsTxt, @tagInnerHTML, @tagOpenCloseMode) ->
    if not @tagInnerHTML?
      @tagInnerHTML = ''
    if not @tagParamsTxt?
      @tagParamsTxt = ''

  tagOpenClose: (tag, txt, params) ->
    rv = @newNate(@)
    rv.setTag tag, params, txt, K.NateHtmlElem.OPEN_THEN_CLOSE
    return rv

  txt: (theTxt) ->
    rv = @newNate(@)
    rv.setTag null, null, theTxt, K.NateHtmlElem.JUST_THE_INNER
    return rv

  newTxt: (theTxt) ->
    return @txt(theTxt)

  br: () ->
    return @txt '<br>'

  newA: (txt, params) ->
    return @tagOpenClose('a', txt, params)

  newB: (txt, params) ->
    return @tagOpenClose('b', txt, params)

  newCanvas: (txt, params) ->
    return @tagOpenClose('canvas', txt, params)

  newDiv: (txt, params) ->
    return @tagOpenClose('div', txt, params)

  newH1: (txt, params) ->
    return @tagOpenClose('h1', txt, params)

  newH2: (txt, params) ->
    return @tagOpenClose('h2', txt, params)

  newH3: (txt, params) ->
    return @tagOpenClose('h3', txt, params)

  newImg: (src, params) ->
    return @tagOpenClose('img', '', 'src="' + src + '" ' + params)

  newInputCheckbox: (txt, params = '') ->
    params = 'type="checkbox" ' + params
    return @tagOpenClose('input', txt, params)

  newInputPassword: (txt, params = '') ->
    params = 'type="password" ' + params
    return @tagOpenClose('input', txt, params)

  newInputRadio: (txt, params = '') ->
    params = 'type="radio" ' + params
    return @tagOpenClose('input', txt, params)

  newInputText: (txt, params = '') ->
    params = 'type="text" ' + params
    return @tagOpenClose('input', txt, params)

  newInputSelect: (txt, params = '') ->
    return @tagOpenClose('select', txt, params)

  newLabel: (txt, params) ->
    return @tagOpenClose('label', txt, params)

  newLi: (txt, params) ->
    return @tagOpenClose('li', txt, params)

  newOption: (txt, params) ->
    return @tagOpenClose('option', txt, params)

  newP: (txt, params) ->
    return @tagOpenClose('p', txt, params)

  newSelect: (txt, params = '') ->
    return @newInputSelect(txt, params)

  newSpan: (txt, params) ->
    return @tagOpenClose('span', txt, params)

  newTextArea: (txt, params) ->
    return @tagOpenClose('textarea', txt, params)

  newUl: (txt, params) ->
    return @tagOpenClose('ul', txt, params)

  newTable: (txt, params) ->
    return @tagOpenClose('table', txt, params)

  newTBody: (txt, params) ->
    return @tagOpenClose('tbody', txt, params)

  newTr: (txt, params) ->
    return @tagOpenClose('tr', txt, params)

  newTd: (txt, params) ->
    return @tagOpenClose('td', txt, params)

  newTh: (txt, params) ->
    return @tagOpenClose('th', txt, params)

  setParams: (params) ->
    @tagParamsTxt = params

  setText: (txt) ->
    @tagInnerHTML = txt
    return @

  setOnClickEventBlocker: () ->
    @set('onclick', K.NateHtmlElem.ON_CLICK_EVENT_BLOCKER)
    return @

  set: (param, value) ->
    switch param
      when 'innerHTML'
        @tagInnerHTML = value
      when 'checked', 'selected'
        if value
          @tagParamsList[param] = param
        else
          if @tagParamsList[param]?
            delete @tagParamsList[param]
      when  'class', 'href', 'id', 'for', 'name', 'onclick', 'style', 'value', \
      'rowSpan', 'colSpan'
        # TODO: value should be string-escaped
        @tagParamsList[param] = value
      when 'on', 'enabled'
        @enabled = value

    return @

  setData: (key, value) ->
    @tagParamsList['data-' + key] = value
    return @

  #
  # Rendering part
  #
  render: () ->
    if @enabled
      switch @tagOpenCloseMode
        when K.NateHtmlElem.JUST_THE_INNER
          rv = @tagInnerHTML
          rv += super()
        when K.NateHtmlElem.OPEN_CLOSE_AT_ONCE, K.NateHtmlElem.OPEN_THEN_CLOSE
          rv = '<' + @tagName
          if @tagParamsTxt isnt ''
            rv += ' ' + @tagParamsTxt
          for paramKey of @tagParamsList
            rv += ' ' + paramKey + '="' + @tagParamsList[paramKey] + '"'
          if @tagOpenClose == K.NateHtmlElem.OPEN_CLOSE_AT_ONCE
            rv += '/>'
          else
            rv += '>'
            rv += super()
            rv += @tagInnerHTML
            rv += '</' + @tagName + '>' + LB
        when K.NateHtmlElem.JUST_CHILDREN
          rv = super()
    else
      rv = ''

    return rv

class K.NateHtmlBody extends K.NateHtmlElem
  constructor: (parent) ->
    super(parent)
    @htmlDocument = parent
    @tagOpenCloseMode = K.NateHtmlElem.JUST_CHILDREN

  render: () ->
    txt = '<body'
    if @tagParamsTxt? and @tagParamsTxt isnt ''
      txt += ' ' + @tagParamsTxt
    txt += '>'
    txt += super()
    txt += @htmlDocument.jsGetCollector('bodyEnd').getAllAsTag()
    txt += @htmlDocument.jsGetCollector('onLoad').getAllAsOnLoadCode()
    txt += '</body>'
    return txt
